import functools
import copy
import random

from cogrid.core import grid_object
from cogrid.envs.overcooked import overcooked
from cogrid.envs import registry
from cogrid.envs.search_rescue import search_rescue
from cogrid.core import layouts


layouts.register_layout(
    "overcooked_cramped_room_v0",
    [
        "#######",
        "#CCUCC#",
        "#O   O#",
        "#C   C#",
        "#C=C@C#",
        "#######",
    ],
)

layouts.register_layout(
    "overcooked_asymmetric_advantages_v0",
    [
        "###########",
        "#CCCCCCCCC#",
        "#O C@COC @#",
        "#C   U   C#",
        "#C   U   C#",
        "#CCC=C=CCC#",
        "###########",
    ],
)

layouts.register_layout(
    "overcooked_coordination_ring_v0",
    [
        "#######",
        "#CCCUC#",
        "#C   U#",
        "#= C C#",
        "#O   C#",
        "#CO@CC#",
        "#######",
    ],
)

layouts.register_layout(
    "overcooked_forced_coordination_v0",
    [
        "#######",
        "#CCCUC#",
        "#O+C U#",
        "#O C C#",
        "#= C+C#",
        "#CCC@C#",
        "#######",
    ],
)

layouts.register_layout(
    "overcooked_counter_circuit_v0",
    [
        "##########",
        "#CCCUUCCC#",
        "#C      C#",
        "#= CCCC @#",
        "#C      C#",
        "#CCCOOCCC#",
        "##########",
    ],
)

cramped_room_config = {
    "name": "overcooked",
    "num_agents": 2,
    "action_set": "cardinal_actions",
    "features": ["overcooked_features"],
    "rewards": ["delivery_reward"],
    "grid": {"layout": "overcooked_cramped_room_v0"},
    "max_steps": 1000,
    "scope": "overcooked",
}

registry.register(
    "Overcooked-CrampedRoom-V0",
    functools.partial(overcooked.Overcooked, config=cramped_room_config),
)

asymmetric_adv_config = copy.deepcopy(cramped_room_config)
asymmetric_adv_config["grid"]["layout"] = "overcooked_asymmetric_advantages_v0"

registry.register(
    "Overcooked-AsymmetricAdvantages-V0",
    functools.partial(overcooked.Overcooked, config=asymmetric_adv_config),
)

coordination_ring_config = copy.deepcopy(cramped_room_config)
coordination_ring_config["grid"]["layout"] = "overcooked_coordination_ring_v0"

registry.register(
    "Overcooked-CoordinationRing-V0",
    functools.partial(overcooked.Overcooked, config=coordination_ring_config),
)

forced_coordination_config = copy.deepcopy(cramped_room_config)
forced_coordination_config["grid"][
    "layout"
] = "overcooked_forced_coordination_v0"

registry.register(
    "Overcooked-ForcedCoordination-V0",
    functools.partial(overcooked.Overcooked, config=forced_coordination_config),
)

counter_circuit_config = copy.deepcopy(cramped_room_config)
counter_circuit_config["grid"]["layout"] = "overcooked_counter_circuit_v0"

registry.register(
    "Overcooked-CounterCircuit-V0",
    functools.partial(overcooked.Overcooked, config=counter_circuit_config),
)


def randomized_layout_fn(**kwargs):
    layout_name = random.choice(
        [
            "overcooked_cramped_room_v0",
            "overcooked_asymmetric_advantages_v0",
            "overcooked_coordination_ring_v0",
            "overcooked_forced_coordination_v0",
            "overcooked_counter_circuit_v0",
        ]
    )
    return layout_name, *layouts.get_layout(layout_name)


overcooked_randomized_config = copy.deepcopy(cramped_room_config)
overcooked_randomized_config["grid"] = {"layout_fn": randomized_layout_fn}

registry.register(
    "Overcooked-RandomizedLayout-V0",
    functools.partial(
        overcooked.Overcooked, config=overcooked_randomized_config
    ),
)


sa_overcooked_config = copy.deepcopy(cramped_room_config)
sa_overcooked_config["num_agents"] = 1
registry.register(
    "Overcooked-CrampedRoom-SingleAgent-V0",
    functools.partial(
        overcooked.Overcooked,
        config=sa_overcooked_config,
        environment_scope="overcooked",
    ),
)


sr_config = {
    "name": "search_rescue",
    "num_agents": 2,
    "action_set": "cardinal_actions",
    "obs": ["agent_positions"],
    "grid": {"layout": "search_rescue_test"},
    "max_steps": 1000,
    "common_reward": True,
    "scope": "search_rescue",
}

layouts.register_layout(
    "search_rescue_test",
    [
        "##########",
        "#++      #",
        "#        #",
        "#        #",
        "#       G#",
        "#        #",
        "#T     K #",
        "#XX M ##D#",
        "#GX Y #GG#",
        "##########",
    ],
)


registry.register(
    "SearchRescue-Test-V0",
    functools.partial(search_rescue.SearchRescueEnv, config=sr_config),
)



import operator

from cogrid.envs.goal_seeking.agent import GoalSeekingAgent
from cogrid.cogrid_env import CoGridEnv


class GoalSeeking(CoGridEnv):
    """
    Goal Seeking GridWorld task. A remake with multi-agent capability. For reference, see:
        https://www.cmu.edu/dietrich/sds/ddmlab/papers/2020ICCM_Ngoc-CameraReady.pdf
        https://www.cmu.edu/dietrich/sds/ddmlab/papers/NguyenGonzalez2020.pdf
    """

    def __init__(self, grid_path, config):
        super().__init__(
            grid_path=grid_path, agent_class=GoalSeekingAgent, config=config
        )

        self.target_values = self.grid_data["values"]
        self.optimal_path_length = self.grid_data["optimal_path_length"]

        self.ma_spawns = (
            [tuple(pos) for pos in self.grid_data["ma_spawns"]]
            if config["num_agents"] > 1
            else None
        )

        self.targets = {}
        for row in range(self.base_grid.shape[0]):
            for col in range(self.base_grid.shape[1]):
                if self.base_grid[row, col] in self.target_values.keys():
                    self.targets[self.base_grid[row, col]] = (row, col)

        self.pref_target_loc = self.targets[
            max(self.target_values.items(), key=operator.itemgetter(1))[0]
        ]

        self.setup_agents()

    def select_spawn_point(self, random_spawn=True) -> tuple:
        curr_pos = [agent.pos for agent in self.env_agents.values()]

        if (
            "gen_random_spawn" in self.config["env"].keys()
            and self.config["env"]["gen_random_spawn"]
        ):
            self.np_random.shuffle(self.free_spaces)
            # note that here, we specify that a random spawn should not be the specified spawn on the map,
            # this behavior may or may not be desired in various experiments, although it is the case
            # for the SoU transfer experiments that we want random positions to be distinct from the pre-defined ones.
            available = [
                sp
                for sp in self.free_spaces
                if sp not in curr_pos and sp not in self.spawns
            ]
            return available[0]

        if random_spawn:
            self.np_random.shuffle(self.spawns)

        if self.config["num_agents"] > 1:
            all_spawns = self.ma_spawns
        else:
            all_spawns = self.spawns

        selected_spawn = None
        for spawn in all_spawns:
            if spawn not in curr_pos:
                selected_spawn = spawn
                break
        assert (
            selected_spawn is not None
        ), "There are not enough spawn points in the map for the specified number of agents."
        return selected_spawn

    def custom_reset(self):
        self.add_goals()
        for agent in self.env_agents.values():
            agent.all_agent_pos = self.agent_pos

    def add_goals(self):
        for target, pos in self.targets.items():
            self.world_map[pos] = target



import dataclasses

import numpy as np

"""
Define all constants for CoGridEnv environments.
"""


@dataclasses.dataclass
class GridConstants:
    FreeSpace = " "
    Padding = "0"
    Obscured = "."
    Spawn = "+"



import numpy as np


LAYOUT_REGISTRY: dict[str, tuple[list[str], list[int]]] = {}


def get_layout(layout_id: str, **kwargs) -> tuple[list[str], list[int]]:
    return LAYOUT_REGISTRY[layout_id]


def register_layout(
    layout_id: str,
    layout: list[str],
    state_encoding: list[list[int]] | np.ndarray | None = None,
) -> None:
    if state_encoding is None:
        state_encoding = np.zeros((len(layout), len(layout[0])))

    if layout_id in LAYOUT_REGISTRY:
        raise ValueError(
            f"There is already a layout registered with ID {layout_id}. Please use a unique ID."
        )

    LAYOUT_REGISTRY[layout_id] = (layout, state_encoding)



import dataclasses

import numpy as np


@dataclasses.dataclass
class CoreConstants:
    TilePixels = 32


@dataclasses.dataclass
class Colors:
    Red = [255, 0, 0]
    Green = [0, 255, 0]
    Blue = [0, 0, 255]
    Purple = [112, 39, 195]
    Yellow = [255, 255, 0]
    Grey = [100, 100, 100]
    DarkGrey = [70, 70, 70]
    LightBrown = [181, 132, 0]
    PukeGreen = [191, 155, 0]
    Brown = [143, 106, 35]
    Cyan = [0, 234, 255]
    Lavender = [220, 185, 237]
    YellowGrey = [231, 233, 185]
    LightPink = [237, 185, 185]
    Orange = [245, 130, 200]
    PaleBlue = [0, 0, 127]
    White = [255, 255, 255]
    Black = [0, 0, 0]


@dataclasses.dataclass
class ObjectColors:

    AgentOne = Colors.Cyan
    AgentTwo = Colors.Lavender
    AgentThree = Colors.YellowGrey
    AgentFour = Colors.PukeGreen


# Map of color names to RGB values
COLORS = {
    "red": np.array([255, 0, 0]),
    "green": np.array([0, 255, 0]),
    "blue": np.array([0, 0, 255]),
    "purple": np.array([112, 39, 195]),
    "yellow": np.array([255, 255, 0]),
    "grey": np.array([100, 100, 100]),
    "dark_grey": np.array([70, 70, 70]),
    "light_brown": np.array([181, 132, 0]),
    "brown": np.array([143, 106, 35]),
    "cyan": np.array([0, 234, 255]),
    "lavender": np.array([220, 185, 237]),
    "yellow_grey": np.array([231, 233, 185]),
    "puke_green": np.array([191, 155, 0]),
    "light_pink": np.array([237, 185, 185]),
    "orange": np.array([245, 130, 200]),
    "pale_blue": np.array([0, 0, 127]),
    "white": np.array([255, 255, 255]),
    "black": np.array([0, 0, 0]),
}

COLOR_NAMES = sorted(list(COLORS.keys()))

# Used to map colors to integers
COLOR_TO_IDX = {c: i for i, c in enumerate(COLORS.keys())}

IDX_TO_COLOR = dict(zip(COLOR_TO_IDX.values(), COLOR_TO_IDX.keys()))



from cogrid.core import reward
from cogrid.core import actions
from cogrid.core.grid import Grid
from cogrid.envs.search_rescue import search_rescue_grid_objects
from cogrid.core import typing


class RescueReward(reward.Reward):
    """Provide a reward for rescuing one of the victims."""

    def __init__(self, agent_ids: list[str | int], **kwargs):
        super().__init__(
            name="rescue_reward", agent_ids=agent_ids, coefficient=1.0, **kwargs
        )

    def calculate_reward(
        self,
        state: Grid,
        agent_actions: dict[typing.AgentID, typing.ActionType],
        new_state: Grid,
    ) -> dict[typing.AgentID, float]:
        """Calcaute the reward for delivering a soup dish.

        :param state: The previous state of the grid.
        :type state: Grid
        :param actions: Actions taken by each agent in the previous state of the grid.
        :type actions: dict[int  |  str, int  |  float]
        :param new_state: The new state of the grid.
        :type new_state: Grid
        """
        prev_num_green = state.get_obj_count(
            search_rescue_grid_objects.GreenVictim
        )
        prev_num_yellow = state.get_obj_count(
            search_rescue_grid_objects.YellowVictim
        )
        prev_num_red = state.get_obj_count(search_rescue_grid_objects.RedVictim)

        new_num_green = new_state.get_obj_count(
            search_rescue_grid_objects.GreenVictim
        )
        new_num_yellow = new_state.get_obj_count(
            search_rescue_grid_objects.YellowVictim
        )
        new_num_red = new_state.get_obj_count(
            search_rescue_grid_objects.RedVictim
        )

        green_reward = prev_num_green - new_num_green
        yellow_reward = (prev_num_yellow - new_num_yellow) * 2
        red_reward = (prev_num_red - new_num_red) * 3

        reward_dict = {
            agent_id: green_reward + yellow_reward + red_reward
            for agent_id in self.agent_ids
        }

        return reward_dict



from cogrid import cogrid_env


ENVIRONMENT_REGISTRY: dict[str, cogrid_env.CoGridEnv] = {}


def make(
    environment_id: str, environment_scope: str = "global", **kwargs
) -> cogrid_env.CoGridEnv:
    if environment_id not in ENVIRONMENT_REGISTRY:
        raise ValueError(
            f"Environment ID {environment_id} is not registered. "
            f"Please register it first. \n "
            f"Available environments: {list(ENVIRONMENT_REGISTRY.keys())}"
        )
    return ENVIRONMENT_REGISTRY[environment_id](
        environment_scope=environment_scope, **kwargs
    )


def register(environment_id: str, env_class: cogrid_env.CoGridEnv) -> None:
    ENVIRONMENT_REGISTRY[environment_id] = env_class



from cogrid.core import grid
from cogrid.core import typing


class Reward:
    """Base class for rewards in CoGrid. Rewards are used to define the environment's reward function
    based on actions and state transitions, e.g., R(s, a, s').
    """

    def __init__(
        self,
        name: str,
        agent_ids: list[typing.AgentID],
        coefficient: float = 1.0,
        common_reward: bool = False,
        **kwargs,
    ):
        """_summary_

        :param name: The reward name to show up in metrics.
        :type name: str
        :param agent_ids: List of agent IDs that this reward applies to.
        :type agent_ids: list[str | int]
        :param coefficient: Scaling coefficient of this reward, defaults to 1.0
        :type coefficient: float, optional
        :param common_reward: Whether or not the reward is given to all agents, defaults to False
        :type common_reward: bool, optional
        """
        self.name = name
        self.coefficient = coefficient
        self.common_reward = common_reward
        self.agent_ids = agent_ids

    def calculate_reward(
        self,
        state: grid.Grid,
        agent_actions: dict[int | str, int | float],
        state_transition: grid.Grid,
    ) -> dict[str | int, float]:
        """Calculates the reward based on the state, actions, and state transition.

        :param state: Previous CoGrid environment state.
        :type state: grid.Grid
        :param agent_actions: Actions taken in the previous state.
        :type agent_actions: dict[int  |  str, int  |  float]
        :param state_transition: Current CoGrid environment state after taking the actions in the previous state.
        :type state_transition: grid.Grid
        :raises NotImplementedError: This method must be implemented in the subclass.
        """
        raise NotImplementedError


REWARD_REGISTRY: dict[str, Reward] = {}


def register_reward(reward_id: str, reward_class: Reward) -> None:
    if reward_id in REWARD_REGISTRY:
        print(
            f"A reward is already registered with the ID {reward_id}. "
            "Overwriting it."
        )

    REWARD_REGISTRY[reward_id] = reward_class


def make_reward(reward_id: str, **kwargs) -> Reward:
    if reward_id not in REWARD_REGISTRY:
        raise ValueError(f"Reward {reward_id} not registered.")
    return REWARD_REGISTRY[reward_id](**kwargs)



import functools

from cogrid.core import reward
from cogrid.core import actions
from cogrid.core.grid import Grid
from cogrid.envs.overcooked import overcooked_grid_objects
from cogrid.core import typing


class SoupDeliveryReward(reward.Reward):
    """Provide a reward for delivery an OnionSoup to a DeliveryZone."""

    def __init__(
        self,
        agent_ids: list[str | int],
        common_reward: bool = True,
        coefficient=1.0,
        **kwargs
    ):
        super().__init__(
            name="delivery_reward",
            agent_ids=agent_ids,
            coefficient=coefficient,
            **kwargs
        )
        self.common_reward = common_reward

    def calculate_reward(
        self,
        state: Grid,
        agent_actions: dict[typing.AgentID, typing.ActionType],
        new_state: Grid,
    ) -> dict[typing.AgentID, float]:
        """Calcaute the reward for delivering a soup dish.

        :param state: The previous state of the grid.
        :type state: Grid
        :param actions: Actions taken by each agent in the previous state of the grid.
        :type actions: dict[int  |  str, int  |  float]
        :param new_state: The new state of the grid.
        :type new_state: Grid
        """
        # Reward is shared among all agents, so calculate once
        # then distribute to all agents

        common_reward = 0
        individual_rewards = {agent_id: 0 for agent_id in self.agent_ids}

        for agent_id, action in agent_actions.items():
            # Check if agent is performing a PickupDrop action
            if action != actions.Actions.PickupDrop:
                continue

            # Check if an agent is holding an OnionSoup
            agent = state.grid_agents[agent_id]
            agent_holding_soup = any(
                [
                    isinstance(obj, overcooked_grid_objects.OnionSoup)
                    for obj in agent.inventory
                ]
            )

            # Check if the agent is facing a delivery zone
            fwd_pos = agent.front_pos
            fwd_cell = state.get(*fwd_pos)
            agent_facing_delivery = isinstance(
                fwd_cell, overcooked_grid_objects.DeliveryZone
            )

            if agent_holding_soup and agent_facing_delivery:
                common_reward += self.coefficient
                individual_rewards[agent_id] += self.coefficient

        if self.common_reward:
            return {agent_id: common_reward for agent_id in self.agent_ids}

        return individual_rewards


reward.register_reward("delivery_reward", SoupDeliveryReward)
reward.register_reward(
    "delivery_reward_individual",
    functools.partial(SoupDeliveryReward, common_reward=False),
)


class OnionInPotReward(reward.Reward):
    """Provide a reward for putting an onion in the pot. This is for reward shaping."""

    def __init__(
        self, agent_ids: list[str | int], coefficient: float = 0.1, **kwargs
    ):
        super().__init__(
            name="onion_in_pot_reward",
            agent_ids=agent_ids,
            coefficient=coefficient,
            **kwargs
        )

    def calculate_reward(
        self,
        state: Grid,
        agent_actions: dict[int | str, int | float],
        new_state: Grid,
    ) -> dict[str | int, float]:
        """Calcaute the reward putting an onion in the pot.

        :param state: The previous state of the grid.
        :type state: Grid
        :param actions: Actions taken by each agent in the previous state of the grid.
        :type actions: dict[int  |  str, int  |  float]
        :param new_state: The new state of the grid.
        :type new_state: Grid
        """
        rewards = {agent_id: 0 for agent_id in self.agent_ids}

        for agent_id, action in agent_actions.items():
            # Check if agent is performing a PickupDrop action
            if action != actions.Actions.PickupDrop:
                continue

            # Check if an agent is holding an Onion
            agent = state.grid_agents[agent_id]
            agent_holding_onion = any(
                [
                    isinstance(obj, overcooked_grid_objects.Onion)
                    for obj in agent.inventory
                ]
            )
            if not agent_holding_onion:
                continue

            # Check if the agent is facing a Pot
            fwd_pos = agent.front_pos
            fwd_cell = state.get(*fwd_pos)
            agent_facing_pot = isinstance(fwd_cell, overcooked_grid_objects.Pot)
            pot_has_capacity = agent_facing_pot and fwd_cell.can_place_on(
                agent=agent, cell=agent.inventory[0]
            )

            if agent_holding_onion and agent_facing_pot and pot_has_capacity:
                rewards[agent_id] = self.coefficient

        return rewards


reward.register_reward("onion_in_pot_reward", OnionInPotReward)


class SoupInDishReward(reward.Reward):
    """Provide a reward for putting the soup into a dish."""

    def __init__(
        self, agent_ids: list[str | int], coefficient: float = 0.3, **kwargs
    ):
        super().__init__(
            name="soup_in_dish_reward",
            agent_ids=agent_ids,
            coefficient=coefficient,
            **kwargs
        )

    def calculate_reward(
        self,
        state: Grid,
        agent_actions: dict[int | str, int | float],
        new_state: Grid,
    ) -> dict[str | int, float]:
        """Calcaute the reward putting the soup .

        :param state: The previous state of the grid.
        :type state: Grid
        :param actions: Actions taken by each agent in the previous state of the grid.
        :type actions: dict[int  |  str, int  |  float]
        :param new_state: The new state of the grid.
        :type new_state: Grid
        """
        rewards = {agent_id: 0 for agent_id in self.agent_ids}

        for agent_id, action in agent_actions.items():
            # Check if agent is performing a PickupDrop action
            if action != actions.Actions.PickupDrop:
                continue

            # Check if an agent is holding a Plate
            agent = state.grid_agents[agent_id]
            agent_holding_soup = any(
                [
                    isinstance(obj, overcooked_grid_objects.Plate)
                    for obj in agent.inventory
                ]
            )

            # Check if the agent is facing a Pot with a ready soup
            fwd_pos = agent.front_pos
            fwd_cell = state.get(*fwd_pos)
            agent_facing_pot = isinstance(fwd_cell, overcooked_grid_objects.Pot)
            facing_pot_and_pot_is_ready = (
                agent_facing_pot and fwd_cell.dish_ready
            )

            if agent_holding_soup and facing_pot_and_pot_is_ready:
                rewards[agent_id] = self.coefficient

        return rewards


reward.register_reward("soup_in_dish_reward", SoupInDishReward)



import numpy as np

from cogrid.constants import GridConstants


def ascii_to_numpy(ascii_list):
    rows, cols = len(ascii_list), len(ascii_list[0])
    for row in range(0, rows):
        assert len(ascii_list[row]) == cols, print("The ascii map is not rectangular!")
    arr = np.full((rows, cols), GridConstants.FreeSpace)
    for row in range(arr.shape[0]):
        for col in range(arr.shape[1]):
            arr[row, col] = ascii_list[row][col]
    return arr


def adjacent_positions(row, col):
    for rdelta, cdelta in [(0, 1), (0, -1), (1, 0), (-1, 0)]:
        yield row + rdelta, col + cdelta



from __future__ import annotations


from cogrid.core.actions import Actions
from cogrid.cogrid_env import CoGridEnv
from cogrid.envs import registry
from cogrid.core import typing


import numpy as np

# try:
#     import onnxruntime as ort
# except ImportError:
#     raise ImportError(
#         "Must `pip install onnxruntime` to use the ONNX inference utils!"
#     )

# try:
#     from scipy import special
# except ImportError:
#     raise ImportError(
#         "Must `pip install scipy` to use the ONNX inference utils!"
#     )


try:
    import pygame
except ImportError:
    pygame = None
    raise ImportError(
        "Must `pip install pygame` to use interactive visualizer!"
    )

ORT_SESSIONS: dict[str, ort.InferenceSession] = {}

REWARDS = []


def sample_action_via_softmax(logits: np.ndarray) -> int:
    """Given logits sample an action via softmax"""
    action_distribution = special.softmax(logits)
    action = np.random.choice(
        np.arange(len(action_distribution)), p=action_distribution
    )
    return action


def inference_onnx_model(
    input_dict: dict[str, np.ndarray],
    model_path: str,
) -> np.ndarray:
    """Given an input dict and path to an ONNX model, return the model outputs"""
    outputs = ORT_SESSIONS[model_path].run(["output"], input_dict)
    return outputs


def onnx_model_inference_fn(
    observation: dict[str, np.ndarray] | np.ndarray, onnx_model_path: str
):
    # if it's a dictionary observation, the onnx model expects a flattened input array
    if isinstance(observation, dict):
        observation = np.hstack(list(observation.values())).reshape((1, -1))

    model_outputs = inference_onnx_model(
        {
            "obs": observation.astype(np.float32),
            "state_ins": np.array([0.0], dtype=np.float32),  # rllib artifact
        },
        model_path=onnx_model_path,
    )[0].reshape(
        -1
    )  # outputs list of a batch. batch size always 1 so index list and reshape

    action = sample_action_via_softmax(model_outputs)

    return action


def load_onnx_policy_fn(onnx_model_path: str) -> str:
    """Initialize the ORT session and return the string to access it"""
    if ORT_SESSIONS.get(onnx_model_path) is None:
        ORT_SESSIONS[onnx_model_path] = ort.InferenceSession(
            onnx_model_path, None
        )

    return onnx_model_path


model = None  # load_onnx_policy_fn("path/to/model.onnx")


ACTION_MESSAGE = ""
HUMAN_AGENT_ID = 1

ACTION_SET = "cardinal_actions"

if ACTION_SET == "cardinal_actions":
    KEY_TO_ACTION = {
        "left": Actions.MoveLeft,
        "right": Actions.MoveRight,
        "up": Actions.MoveUp,
        "down": Actions.MoveDown,
        "q": Actions.Toggle,
        "w": Actions.PickupDrop,
        "space": Actions.Noop,
    }
elif ACTION_SET == "rotation_actions":
    KEY_TO_ACTION = {
        "left": Actions.RotateLeft,
        "right": Actions.RotateRight,
        "up": Actions.Forward,
        "q": Actions.Toggle,
        "w": Actions.PickupDrop,
        "space": Actions.Noop,
    }


class HumanPlay:
    def __init__(
        self,
        env: CoGridEnv,
        human_agent_id: typing.AgentID = None,
        seed: int = None,
    ) -> None:
        self.env = env
        self.seed = seed
        self.closed = False
        self.human_agent_id = human_agent_id
        self.obs = None
        self.cumulative_reward = 0

    def run(self):
        self.reset(self.seed)
        while not self.closed:
            actions = {
                agent_id: Actions.Noop for agent_id in self.env.agent_ids
            }

            for a_id, obs in self.obs.items():
                if a_id == self.human_agent_id:
                    continue
                actions[a_id] = self.env.action_spaces[a_id].sample()
                # if self.env.t % 5 == 0:
                #     actions[a_id] = onnx_model_inference_fn(obs, model)

            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    self.env.close()
                    return
                if event.type == pygame.KEYDOWN:
                    event.key = pygame.key.name((int(event.key)))

                    if event.key == "escape":
                        self.env.close()
                        return

                    if event.key == "backspace":
                        self.reset(self.seed)
                        return

                    if (
                        self.human_agent_id is not None
                        and event.key in KEY_TO_ACTION.keys()
                    ):
                        actions[self.human_agent_id] = KEY_TO_ACTION[event.key]
                    else:
                        print(f"Invalid action: {event.key}")

            self.step(actions)

    def step(self, actions: dict[str, Actions]):
        self.obs, rewards, terminateds, truncateds, _ = self.env.step(actions)
        self.cumulative_reward += [*rewards.values()][0]
        # print(
        #     f"step={self.env.t}, rewards={rewards}, cumulative_reward={self.cumulative_reward}"
        # )

        if not self.env.agents:
            print("All agents done!")
            REWARDS.append(self.cumulative_reward)
            if len(REWARDS) == 50:
                print(REWARDS)
            self.reset(self.seed)
        else:
            self.env.render()

    def reset(self, seed):
        self.obs, _ = self.env.reset(seed=seed)
        self.cumulative_reward = 0
        self.env.render()


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--env-id",
        type=str,
        help="environment to load",
        default="search_rescue",
    )
    parser.add_argument(
        "--seed",
        type=int,
        help="random seed to generate the environment with",
        default=42,
    )
    parser.add_argument(
        "--tile-size",
        type=int,
        help="size at which to render tiles",
        default=32,
    )
    parser.add_argument(
        "--agent-pov",
        action="store_true",
        help="draw the agent sees (partially observable view)",
    )
    parser.add_argument(
        "--agent-view-size",
        type=int,
        default=7,
        help="set the number of grid spaces visible in agent-view ",
    )
    parser.add_argument(
        "--screen-size",
        type=int,
        default=512,
        help="set the resolution for pygame rendering (width and height)",
    )

    args = parser.parse_args()

    def env_creator(
        render_mode: str | None = None, render_message=""
    ) -> CoGridEnv:
        return registry.make(
            "Overcooked-ForcedCoordination-V0",
            highlight=False,
            render_mode=render_mode,
            screen_size=args.screen_size,
            render_message=render_message,
        )

    policy_mapping = {0: "random", 1: "random"}

    # NOTE: If you need to pass a config to your policy, specify it here and the
    # policy class will be initialized with it.
    configs = {agent_id: {} for agent_id in policy_mapping.keys()}

    env: CoGridEnv = env_creator(render_mode="human")
    manual_control = HumanPlay(
        env,
        human_agent_id=HUMAN_AGENT_ID,
        seed=args.seed,
    )
    manual_control.run()



from cogrid import cogrid_env
from cogrid.envs.search_rescue import search_rescue_grid_objects
from cogrid.envs import registry


class SearchRescueEnv(cogrid_env.CoGridEnv):
    """
    The Search & Rescue task is a reproduction of the __Minimap__ game.
    """

    def get_terminateds_truncateds(self) -> tuple:
        """
        returns dones only when all targets have been located.
        """
        green_targets_in_grid = any(
            [
                isinstance(obj, search_rescue_grid_objects.GreenVictim)
                for obj in self.grid.grid
            ]
        )
        yellow_targets_in_grid = any(
            [
                isinstance(obj, search_rescue_grid_objects.YellowVictim)
                for obj in self.grid.grid
            ]
        )
        red_targets_in_grid = any(
            [
                isinstance(obj, search_rescue_grid_objects.RedVictim)
                for obj in self.grid.grid
            ]
        )

        all_targets_reached = (
            not green_targets_in_grid
            and not yellow_targets_in_grid
            and not red_targets_in_grid
        )

        if all_targets_reached:
            for agent in self.env_agents.values():
                agent.terminated = True

        return super().get_terminateds_truncateds()


registry.register("search_rescue", SearchRescueEnv)



from cogrid.core import grid_object

from cogrid.constants import GridConstants
from cogrid.core.roles import Roles
from cogrid.core.constants import ObjectColors, COLORS
from cogrid.core import constants
from cogrid.core.grid_utils import adjacent_positions
from cogrid.core import typing
from cogrid.visualization.rendering import (
    fill_coords,
    point_in_circle,
    point_in_rect,
    point_in_triangle,
)


class MedKit(grid_object.GridObj):
    object_id = "medkit"
    color = constants.Colors.LightPink
    char = "M"

    def __init__(self, state=0):
        super().__init__(
            state=state,
        )

    def can_pickup(self, agent: grid_object.GridAgent):
        return True

    def render(self, tile_img):
        # red background with white cross
        fill_coords(
            tile_img, point_in_rect(0.1, 0.9, 0.1, 0.9), (255, 0, 0)
        )  # red background
        fill_coords(
            tile_img, point_in_rect(0.4, 0.6, 0.2, 0.8), (255, 255, 255)
        )  # vertical bar
        fill_coords(
            tile_img, point_in_rect(0.2, 0.8, 0.4, 0.6), (255, 255, 255)
        )  # horizontal bar


grid_object.register_object(MedKit.object_id, MedKit, scope="search_rescue")


class Pickaxe(grid_object.GridObj):
    object_id = "pickaxe"
    color = constants.Colors.Grey
    char = "T"

    def __init__(self, state=0):
        super().__init__(
            state=state,
        )

    def can_pickup(self, agent: grid_object.GridAgent):
        return True

    def render(self, tile_img):

        # Brown Handle
        fill_coords(
            tile_img, point_in_rect(0.45, 0.55, 0.15, 0.9), constants.Colors.Brown
        )

        # Use two triangles to make the pickaxe head
        # These are of the specified color
        tri_fn = point_in_triangle(
            (0.5, 0.1),
            (0.5, 0.3),
            (0.9, 0.35),
        )
        fill_coords(tile_img, tri_fn, self.color)

        tri_fn = point_in_triangle(
            (0.5, 0.1),
            (0.5, 0.3),
            (0.1, 0.35),
        )
        fill_coords(tile_img, tri_fn, self.color)


grid_object.register_object(Pickaxe.object_id, Pickaxe, scope="search_rescue")


class Rubble(grid_object.GridObj):
    object_id = "rubble"
    color = constants.Colors.Brown
    char = "X"

    def __init__(self, state=0):
        super().__init__(
            state=state,
            toggle_value=0.05,  # reward for clearing rubble
        )

    def see_behind(self) -> bool:
        return False

    def toggle(self, env, agent=None) -> bool:
        """Rubble can be toggled by an Engineer/agent with Pickaxe"""
        assert agent
        adj_positions = [*adjacent_positions(*self.pos)]
        toggling_agent_is_adjacent = tuple(agent.pos) in adj_positions
        toggling_agent_is_engineer = (
            any([isinstance(obj, Pickaxe) for obj in agent.inventory])
            or agent.role == Roles.Engineer
        )

        assert toggling_agent_is_adjacent, "Rubble toggled by non-adjacent agent."

        toggle_success = toggling_agent_is_engineer

        if toggle_success:
            self._remove_from_grid(env.grid)

        return toggle_success

    def render(self, tile_img):
        fill_coords(tile_img, point_in_circle(cx=0.25, cy=0.3, r=0.2), self.color)
        fill_coords(tile_img, point_in_circle(cx=0.75, cy=0.3, r=0.2), self.color)
        fill_coords(tile_img, point_in_circle(cx=0.5, cy=0.7, r=0.2), self.color)


grid_object.register_object(Rubble.object_id, Rubble, scope="search_rescue")


class GreenVictim(grid_object.GridObj):
    object_id = "green_victim"
    color = constants.Colors.Green
    char = "G"

    def __init__(self, state=0):
        super().__init__(
            state=state,
            toggle_value=0.1,  # 0.1 reward for rescuing
        )

    def toggle(self, env, agent=None) -> bool:
        """Toggling a victim rescues them. A GreenVictim can be rescued if any agent is adjacent to it"""
        # Toggle should only be triggered if the GreenVictim is directly in front of it. For debugging purposes,
        # we'll just check to make sure that's true (if this isn't triggered it can be removed).
        assert agent
        adj_positions = [*adjacent_positions(*self.pos)]
        toggling_agent_is_adjacent = tuple(agent.pos) in adj_positions
        assert toggling_agent_is_adjacent, "GreenVictim toggled by non-adjacent agent."

        self._remove_from_grid(env.grid)
        return toggling_agent_is_adjacent

    def render(self, tile_img):
        fill_coords(tile_img, point_in_circle(cx=0.5, cy=0.47, r=0.4), self.color)


grid_object.register_object(GreenVictim.object_id, GreenVictim, scope="search_rescue")


class PurpleVictim(grid_object.GridObj):
    object_id = "purple_victim"
    color = constants.Colors.Purple
    char = "P"

    def __init__(self, state=0):
        super().__init__(
            state=state,
            toggle_value=0.2,
        )

    def toggle(self, env, agent=None) -> bool:
        """Toggling a victim rescues them. A PurpleVictim can be rescued if any agent is adjacent to it"""
        # Toggle should only be triggered if the GreenVictim is directly in front of it. For debugging purposes,
        # we'll just check to make sure that's true (if this isn't triggered it can be removed).
        assert agent
        adj_positions = [*adjacent_positions(*self.pos)]
        toggling_agent_is_adjacent = tuple(agent.pos) in adj_positions
        assert toggling_agent_is_adjacent, "GreenVictim toggled by non-adjacent agent."

        self._remove_from_grid(env.grid)
        return toggling_agent_is_adjacent

    def render(self, tile_img):
        fill_coords(tile_img, point_in_circle(cx=0.5, cy=0.47, r=0.4), self.color)


grid_object.register_object(PurpleVictim.object_id, PurpleVictim, scope="search_rescue")


class YellowVictim(grid_object.GridObj):
    object_id = "yellow_victim"
    color = constants.Colors.Yellow
    char = "Y"

    def __init__(self, state=0):
        super().__init__(
            state=state,
            toggle_value=0.2,
        )

    def toggle(self, env, agent=None) -> bool:
        """
        Toggling a victim rescues them. A YellowVictim can be rescued if a Medic is adjcent to it or the agent
        is carrying a MedKit
        """
        assert agent
        adj_positions = [*adjacent_positions(*self.pos)]
        toggling_agent_is_adjacent = tuple(agent.pos) in adj_positions
        toggling_agent_is_medic = (
            any([isinstance(obj, MedKit) for obj in agent.inventory])
            or agent.role == Roles.Medic
        )

        assert toggling_agent_is_adjacent, "YellowVictim toggled by non-adjacent agent."

        toggle_success = toggling_agent_is_medic

        if toggle_success:
            self._remove_from_grid(env.grid)
        return toggle_success

    def render(self, tile_img):
        fill_coords(tile_img, point_in_circle(cx=0.5, cy=0.47, r=0.4), self.color)


grid_object.register_object(YellowVictim.object_id, YellowVictim, scope="search_rescue")


class RedVictim(grid_object.GridObj):
    object_id = "red_victim"
    color = constants.Colors.Red
    char = "R"

    def __init__(self, state=0):
        super().__init__(
            state=state,
        )
        self.toggle_countdown = 0
        self.first_toggle_agent_id: typing.AgentID = None

    def tick(self):
        """At each timestep, decrement toggle countdown and set the count as the state."""
        if self.toggle_countdown > 0:
            self.toggle_countdown -= 1
        self.state = self.toggle_countdown

    def toggle(self, env, agent) -> bool:
        """A RedVictim can be rescued if a Medic (or agent carrying MedKit) is the adjacent toggling agent
        and then another agent toggles within 30 timesteps."""

        if self.toggle_countdown == 0:

            toggling_agent_has_medkit = any(
                [isinstance(obj, MedKit) for obj in agent.inventory]
            )

            if toggling_agent_has_medkit:
                self.first_toggle_agent = agent.agent_id
                self.toggle_countdown = 30

            return True

        if self.toggle_countdown > 0 and agent.agent_id != self.first_toggle_agent:
            self._remove_from_grid(env.grid)
            return True

        return False

    def render(self, tile_img):
        fill_coords(tile_img, point_in_circle(cx=0.5, cy=0.47, r=0.4), self.color)


grid_object.register_object(RedVictim.object_id, RedVictim, scope="search_rescue")



import numpy as np
from gymnasium import spaces

from cogrid.core import typing


class Feature:
    """Base class for creating observation features.

    :param low: Minimum value for this feature.
    :type low: float
    :param high: Maximum value for this feature.
    :type high: float
    :param name: Descriptive name for this feature.
    :type name: str
    :param shape: Shape of the feature, defaults to None
    :type shape: tuple | np.ndarray, optional
    :param space: Gymnasium Space for this features, defaults to None
    :type space: spaces.Space | None, optional
    """

    shape = None

    def __init__(
        self,
        low: float,
        high: float,
        name: str,
        shape: tuple | np.ndarray = None,
        space: spaces.Space | None = None,
        **kwargs
    ):
        """Generator Method"""
        self.low = low
        self.high = high

        if shape is not None:
            self.shape = shape
        assert self.shape is not None, "Must specify shape via class or init!"

        self.space = spaces.Box(low, high, shape) if space is None else space
        self.name = name

    def generate(
        self, env: typing.EnvType, player_id: typing.AgentID, **kwargs
    ):
        """Generate the feature for the specified player in environment.

        :param env: The CoGridEnv instance to calculate the feature from (e.g., state S).
        :type env: cogrid_env.CoGridEnv
        :param player_id: The AgentID to calculate the feature for.
        :type player_id: typing.AgentID
        :raises NotImplementedError: Must be implemented by subclass.
        """
        raise NotImplementedError



import setuptools

setuptools.setup(
    name="cogrid",
    version="0.0.11",
    description="Multi-agent extension and expansion of MiniGrid.",
    author="Chase McDonald",
    author_email="chasemcd@cmu.edu",
    url="https://github.com/chasemcd/cogrid",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
    install_requires=[
        "numpy==1.26.4",
        "gymnasium==1.0.0",
        "pettingzoo==1.24.3",
    ],
)



from __future__ import annotations
from copy import deepcopy
import math

import numpy as np

from cogrid.constants import GridConstants
from cogrid.core.constants import COLORS
from cogrid.core import constants
from cogrid.visualization.rendering import (
    fill_coords,
    point_in_circle,
    point_in_rect,
    point_in_triangle,
    rotate_fn,
    add_text_to_image,
)

from cogrid.core import grid_object


class Onion(grid_object.GridObj):
    object_id = "onion"
    color = constants.Colors.Yellow
    char = "o"

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(
            state=0,
            inventory_value=0.0,
        )

    def can_pickup(self, agent: grid_object.GridAgent) -> bool:
        return True

    def render(self, tile_img):
        fill_coords(
            tile_img, point_in_circle(cx=0.5, cy=0.5, r=0.3), self.color
        )


grid_object.register_object(Onion.object_id, Onion, scope="overcooked")


class OnionStack(grid_object.GridObj):
    """An OnionStack is just an (infinite) pile of onions."""

    object_id = "onion_stack"
    color = constants.Colors.Yellow
    char = "O"

    def can_pickup_from(self, agent: grid_object.GridAgent) -> bool:
        return True

    def pick_up_from(self, agent: grid_object.GridAgent) -> grid_object.GridObj:
        return Onion()

    def render(self, tile_img: np.ndarray):
        fill_coords(
            tile_img, point_in_circle(cx=0.25, cy=0.3, r=0.2), self.color
        )
        fill_coords(
            tile_img, point_in_circle(cx=0.75, cy=0.3, r=0.2), self.color
        )
        fill_coords(
            tile_img, point_in_circle(cx=0.5, cy=0.7, r=0.2), self.color
        )


grid_object.register_object(
    OnionStack.object_id, OnionStack, scope="overcooked"
)


class Pot(grid_object.GridObj):
    object_id = "pot"
    color = constants.Colors.Grey
    char = "U"
    cooking_time: int = 30  # env steps to cook a soup

    def __init__(
        self,
        state: int = 0,
        capacity: int = 3,
        legal_contents: list[grid_object.GridObj] = [Onion],
        *args,
        **kwargs,
    ):

        super().__init__(
            state=state, picked_up_from_value=0.0, placed_on_value=0.0
        )

        self.objects_in_pot: list[grid_object.GridObj] = []
        self.capacity: int = capacity
        self.cooking_timer: int = self.cooking_time
        self.legal_contents: list[grid_object.GridObj] = legal_contents

    def can_pickup_from(self, agent: grid_object.GridAgent) -> bool:
        return self.dish_ready and any(
            [isinstance(grid_obj, Plate) for grid_obj in agent.inventory]
        )

    def pick_up_from(self, agent: grid_object.GridAgent) -> grid_object.GridObj:
        self.objects_in_pot = []
        self.cooking_timer = self.cooking_time
        agent.inventory.pop(0)  # TODO(chase): assumes size 1 inventory
        return OnionSoup()

    def can_place_on(
        self, agent: grid_object.GridAgent, cell: grid_object.GridObj
    ) -> bool:
        """Can only place onions in the soup!"""
        if not any(
            [isinstance(cell, grid_obj) for grid_obj in self.legal_contents]
        ):
            return False

        return len(self.objects_in_pot) < self.capacity

    def place_on(
        self, agent: grid_object.GridAgent, cell: grid_object.GridObj
    ) -> None:
        self.objects_in_pot.append(cell)

    @property
    def is_cooking(self) -> None:
        return (
            len(self.objects_in_pot) == self.capacity and self.cooking_timer > 0
        )

    def tick(self) -> None:
        """Update cooking time if the pot is full"""
        if len(self.objects_in_pot) == self.capacity and self.cooking_timer > 0:
            self.cooking_timer -= 1
            self.state += 100

        self.state = (
            len(self.objects_in_pot)
            + len(self.objects_in_pot) * self.cooking_timer
        )

    @property
    def dish_ready(self) -> bool:
        return self.cooking_timer == 0

    def render(self, tile_img):
        fill_coords(
            tile_img, point_in_circle(cx=0.5, cy=0.5, r=0.5), self.color
        )

        for i, grid_obj in enumerate(self.objects_in_pot):
            fill_coords(
                tile_img,
                point_in_circle(cx=0.25 * (i + 1), cy=0.2, r=0.2),
                grid_obj.color,
            )

        if len(self.objects_in_pot) == self.capacity:
            add_text_to_image(
                tile_img, text=str(self.cooking_timer), position=(50, 75)
            )


grid_object.register_object(Pot.object_id, Pot, scope="overcooked")


class PlateStack(grid_object.GridObj):
    object_id = "plate_stack"
    color = constants.Colors.White
    char = "="

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(
            state=0,
            toggle_value=0,
            inventory_value=0,
            overlap_value=0,
        )

    def can_pickup_from(self, agent: grid_object.GridAgent) -> bool:
        return True

    def pick_up_from(self, agent: grid_object.GridAgent) -> grid_object.GridObj:
        return Plate()

    def render(self, tile_img):
        fill_coords(
            tile_img, point_in_circle(cx=0.25, cy=0.3, r=0.2), self.color
        )
        fill_coords(
            tile_img, point_in_circle(cx=0.75, cy=0.3, r=0.2), self.color
        )
        fill_coords(
            tile_img, point_in_circle(cx=0.5, cy=0.7, r=0.2), self.color
        )


grid_object.register_object(
    PlateStack.object_id, PlateStack, scope="overcooked"
)


class Plate(grid_object.GridObj):
    object_id = "plate"
    color = constants.Colors.White
    char = "P"

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(
            state=0,
            toggle_value=0,
            inventory_value=0.0,
            overlap_value=0,
        )

    def can_pickup(self, agent: grid_object.GridAgent) -> bool:
        return True

    def render(self, tile_img):
        fill_coords(
            tile_img, point_in_circle(cx=0.5, cy=0.5, r=0.5), self.color
        )


grid_object.register_object(Plate.object_id, Plate, scope="overcooked")


class DeliveryZone(grid_object.GridObj):
    object_id = "delivery_zone"
    color = constants.Colors.Green
    char = "@"

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(state=0, toggle_value=0.0, placed_on_value=0.0)

    def can_place_on(
        self, agent: grid_object.GridAgent, cell: grid_object.GridObj
    ) -> bool:
        """Delivery can be toggled by an agent with Soup"""
        toggling_agent_has_soup = any(
            [isinstance(grid_obj, OnionSoup) for grid_obj in agent.inventory]
        )

        if toggling_agent_has_soup:
            return True

        return False

    def place_on(
        self, agent: grid_object.GridAgent, cell: grid_object.GridObj
    ) -> None:
        del cell


grid_object.register_object(
    DeliveryZone.object_id, DeliveryZone, scope="overcooked"
)


class OnionSoup(grid_object.GridObj):
    object_id = "onion_soup"
    color = constants.Colors.LightBrown
    char = "S"

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(
            state=0,
            inventory_value=0.0,
        )

    def can_pickup(self, agent: grid_object.GridAgent) -> bool:
        return True

    def render(self, tile_img):
        # Draw plate
        fill_coords(
            tile_img, point_in_circle(cx=0.5, cy=0.5, r=0.5), Plate.color
        )

        # draw soup inside plate
        fill_coords(
            tile_img,
            point_in_circle(cx=0.5, cy=0.5, r=0.3),
            self.color,
        )


grid_object.register_object(OnionSoup.object_id, OnionSoup, scope="overcooked")



from gymnasium.spaces import Dict

from cogrid.feature_space import feature

FEATURE_SPACE_REGISTRY: dict[str, feature.Feature] = {}


def register_feature(feature_id: str, feature_class: feature.Feature) -> None:
    if feature_id in FEATURE_SPACE_REGISTRY:
        print(
            f"A feature is already registered with the ID `{feature_id}`. Overwriting it."
        )

    FEATURE_SPACE_REGISTRY[feature_id] = feature_class


def make_feature_generator(feature_id: str, **kwargs) -> feature.Feature:
    if feature_id not in FEATURE_SPACE_REGISTRY:
        raise ValueError(f"Feature {feature_id} not registered.")
    return FEATURE_SPACE_REGISTRY[feature_id](**kwargs)


class FeatureSpace:
    def __init__(self, feature_names: list[str], env, agent_id: str, **kwargs):
        self.env = env
        self.feature_generators = self._setup_feature_generators(feature_names)
        self.observation_space = Dict(
            {feature.name: feature.space for feature in self.feature_generators}
        )
        self.agent_id = agent_id

    def generate_features(self):
        generated_features = {
            feature.name: feature.generate(self.env, self.agent_id)
            for feature in self.feature_generators
        }
        return generated_features

    def _setup_feature_generators(self, feature_names):
        return [
            self._fetch_feature_generator(feature_name, env=self.env)
            for feature_name in feature_names
        ]

    def _fetch_feature_generator(self, feat_name, **kwargs):
        return make_feature_generator(feature_id=feat_name, **kwargs)

        # TODO(chase): all of these features should have args replaced with env
        # if feat_name == "full_map_image":
        #     return features.FullMapImage(self.env.shape, self.env.tile_size)
        # if feat_name == "fov_image":
        #     return features.FoVImage(self.env.agent_view_size, self.env.tile_size)
        # if feat_name == "full_map_encoding":
        #     return features.FullMapEncoding(self.env.shape)
        # if feat_name == "fov_encoding":
        #     return features.FoVEncoding(self.env.agent_view_size)
        # if feat_name == "full_map_ascii":
        #     return features.FullMapASCII(self.env.shape)
        # if feat_name == "fov_ascii":
        #     return features.FoVASCII(self.env.agent_view_size)
        # if feat_name == "agent_positions":
        #     return features.AgentPositions(self.env.shape)
        # if feat_name == "other_agent_actions":
        #     return features.OtherAgentActions(
        #         len(self.env.agent_ids), self.env.action_space.n
        #     )
        # if feat_name == "inventory":
        #     # TODO(chase): If functionality added for >1 inventory item, must change this.
        #     return features.Inventory(inventory_capacity=1)
        # if feat_name == "other_agent_visibility":
        #     return features.OtherAgentVisibility(
        #         len(self.env.agent_ids), self.env.agent_view_size
        #     )
        # if feat_name == "role":
        #     return features.Role(self.env.num_roles)
        # if feat_name == "action_mask":
        #     return features.ActionMask(self.env.action_space.n)
        # if feat_name == "agent_id":
        #     return features.AgentID(len(self.env.agent_ids))
        # if feat_name == "stacked_full_map_resized_grayscale_image":
        #     return features.StackedFullMapResizedGrayscale()
        # if feat_name == "full_map_resized_grayscale_image":
        #     return features.FullMapResizedGrayscale()
        # raise NotImplementedError(f"Feature {feat_name} not implemented.")
